using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

using gudusoft.gsqlparser;
using gudusoft.gsqlparser.Units;

namespace TableColumnRef
{
    #region SqlFileList
    public class sqlFileList
    {

        public List<FileInfo> sqlfiles = new List<FileInfo>();

        public sqlFileList(string pattern)
        {
            if (pattern == null)
                return;
            string patternString = pattern.Replace('/', '\\');
            int index = patternString.LastIndexOf('\\') + 1;
            string parentPath = "";
            string filePattern = pattern;
            if (index > 0)
            {
                parentPath = patternString.Substring(0, index);
                filePattern = patternString.Substring(index, patternString.Length - parentPath.Length);
            }
            DirectoryInfo parent = new DirectoryInfo(parentPath);
            if (parent.Exists)
                traverseSqlFiles(parent, filePattern);
        }

        private void traverseSqlFiles(DirectoryInfo parent, string filePattern)
        {
            if (parent == null || !parent.Exists || filePattern == null) return;
            String[] splits = filePattern.Split('*');
            StringBuilder buffer = new StringBuilder();
            buffer.Append(@"^(?i)");
            for (int i = 0; i < splits.Length; i++)
            {
                buffer.Append(@splits[i]);
                if (i < splits.Length - 1)
                {
                    buffer.Append(".*");
                }
            }
            if (filePattern.EndsWith("*") && !buffer.ToString().EndsWith("*"))
                buffer.Append(".*");
            buffer.Append(@"$");
            string pattern = buffer.ToString();
            FileSystemInfo[] files = parent.GetFileSystemInfos();
            for (int i = 0; i < files.Length; i++)
            {
                if (files[i] is FileInfo && Regex.IsMatch(files[i].Name, pattern))
                {
                    sqlfiles.Add((FileInfo)files[i]);
                }
            }
        }
    }
    #endregion

    #region tableInfo
    class tableInfo
    {
        public TSourceToken schema, table;

        public string getSchemaName()
        {
            if (schema == null)
            {
                return "NULL";
            }
            else
            {
                return schema.AsText.Trim();
            }
        }

        public string getTableName()
        {
            if (table == null)
            {
                return "NULL";
            }
            else
            {
                return table.AsText.Trim();
            }
        }
    }

    #endregion

    #region columnInfo
    class columnInfo
    {
        public tableInfo table;
        public TSourceToken column;
        public columnInfo(tableInfo t)
        {
            table = new tableInfo();
            table.schema = t.schema;
            table.table = t.table;
        }
    }

    #endregion

    public class tableColumnRef
    {
        static void Main(string[] args)
        {

            if (args.Length == 0)
            {
                Console.WriteLine("Usage: {0} scriptfile [/o <output file path>] [/t <database type>]", Environment.GetCommandLineArgs()[0]);
                Console.WriteLine("/o: Option, write the output stream to the specified file.");
                Console.WriteLine("/t: Option, set the database type. Support oracle,mssql,mysql and db2, the default type is oracle");
                //Console.Read();
                return;
            }


            Array array = Array.CreateInstance(typeof(string), args.Length);

            for (int i = 0; i < array.Length; i++)
            {
                array.SetValue(args[i], i);
            }

            string outputFile = null;

            int index = Array.IndexOf(array, "/o");

            if (index != -1 && args.Length > index + 1)
            {
                outputFile = args[index + 1];
            }

            TDbVendor vendor = TDbVendor.DbVOracle;

            index = Array.IndexOf(array, "/t");

            if (index != -1 && args.Length > index + 1)
            {
                if (args[index + 1].Equals("mssql"))
                {
                    vendor = TDbVendor.DbVMssql;
                }
                else if (args[index + 1].Equals("db2"))
                {
                    vendor = TDbVendor.DbVDB2;
                }
                else if (args[index + 1].Equals("mysql"))
                {
                    vendor = TDbVendor.DbVMysql;
                }
            }

            tableColumnRef impact = new tableColumnRef(args[0], vendor);

            StreamWriter writer = null;
            if (outputFile != null)
            {
                try
                {
                    writer = new StreamWriter(outputFile);
                    Console.SetOut(writer);
                }
                finally { }
            }

            Console.WriteLine("QUERY_NM,SCHEMA_NM,TABLE_NM,COLUMN_NM,LINE_NR,POS_NR,DETERMINED");
            Console.Write(impact.RefResult.ToUpper());

            //if (outputFile == null)
            //{
            //    Console.Read();
            //}
            //else
            {
                if (writer != null)
                {
                    writer.Close();
                }
            }
        }


        List<columnInfo> columnInfos = null;
        Hashtable columnMap = new Hashtable();
        StringBuilder buffer = new StringBuilder();
        List<string> files = new List<string>();

        public tableColumnRef(string sqlFile, TDbVendor dbVendor)
        {
            sqlFileList sqlFiles = new sqlFileList(sqlFile);
            if (sqlFiles.sqlfiles.Count > 0)
            {
                for (int i = 0; i < sqlFiles.sqlfiles.Count; i++)
                {
                    TGSqlParser sqlparser = new TGSqlParser(dbVendor);
                    sqlparser.Sqlfilename = sqlFiles.sqlfiles[i].FullName;
                    int ret = sqlparser.Parse();
                    if (ret != 0)
                    {
                        Console.WriteLine("Parse file " + sqlparser.Sqlfilename + " failed.");
                        Console.WriteLine(sqlparser.ErrorMessages);
                        continue;
                    }
                    columnInfos = new List<columnInfo>();
                    files.Add(sqlFiles.sqlfiles[i].Name);
                    columnMap.Add(sqlFiles.sqlfiles[i].Name, columnInfos);
                    analyzeSQL(sqlparser);


                    columnInfos.Sort(delegate(columnInfo a, columnInfo b)
                    {
                        int xdiff = a.column.XPosition - b.column.XPosition;
                        if (xdiff != 0) return xdiff;
                        return a.column.YPosition - b.column.YPosition;
                    });
                }

                foreach (string file in files)
                {
                    List<columnInfo> columns = (List<columnInfo>)columnMap[file];
                    List<string> tables = new List<string>();
                    Hashtable resultMap = new Hashtable();
                    foreach (columnInfo info in columns)
                    {
                        string table = (info.table.getSchemaName() + "," + info.table.getTableName()).ToLower().Replace("null", "NULL");
                        if (!resultMap.ContainsKey(table))
                        {
                            resultMap.Add(table, new List<columnInfo>());
                            tables.Add(table);
                        }

                        if (!info.column.AsText.Trim().Equals("*") && !info.column.AsText.Trim().Equals("null", StringComparison.OrdinalIgnoreCase))
                        {
                            List<columnInfo> list = (List<columnInfo>)resultMap[table];
                            bool flag = false;
                            foreach (columnInfo temp in list)
                            {
                                if (temp.column.AsText.Trim().Equals(info.column.AsText.Trim(), StringComparison.OrdinalIgnoreCase))
                                {
                                    flag = true;
                                    break;
                                }
                            }
                            if (!flag)
                            {
                                list.Add(info);
                            }
                        }
                    }

                    foreach (string table in tables)
                    {
                        List<columnInfo> list = (List<columnInfo>)resultMap[table];
                        if (list.Count == 0)
                        {
                            buffer.Append(file).Append(",").Append(table).Append(",")
                                .Append("NULL").Append(",").Append("0")
                                .Append(",").Append("0").Append("no").AppendLine();
                        }
                        else
                        {
                            foreach (columnInfo info in list)
                            {
                                buffer.Append(file).Append(",")
                                       .Append(info.table.getSchemaName()).Append(",")
                                       .Append(info.table.getTableName()).Append(",")
                                       .Append(info.column.AsText.Trim()).Append(",")
                                       .Append(info.column.XPosition).Append(",")
                                       .Append(info.column.YPosition).Append(",")
                                       .Append(info.column.isTableDetermined?"yes":"no").AppendLine();
                            }
                        }
                    }

                }
            }
        }


        public String RefResult
        {
            get { return buffer.ToString(); }
        }

        void analyzeSQL(TGSqlParser sqlparser)
        {

            for (int i = 0; i < sqlparser.SqlStatements.Count(); i++)
            {
                TCustomSqlStatement sql = sqlparser.SqlStatements[i];
                tableTokensInStmt(sql, 0);

                for (int j = 0; j < sql.ChildNodes.Count(); j++)
                {
                    if (sql.ChildNodes[j] is TCustomSqlStatement)
                    {
                        tablesInStmt(sql.ChildNodes[j] as TCustomSqlStatement, 0);
                    }
                }
            }
        }

        void tableTokensInStmt(TCustomSqlStatement stmt, int level)
        {
            TSourceTokenList tokenList = stmt.TableTokens;

            TSourceToken st = null;

            for (int i = 0; i < tokenList.Count(); i++)
            {
                tableInfo tableInfo = new tableInfo();
                st = tokenList[i];
                tableInfo.table = st;
                if (st.ParentToken != null)
                {
                    //schema
                    //tablestr = st.ParentToken.AsText + "." + tablestr;
                    tableInfo.schema = st.ParentToken;
                }
                columnsInTableToken(stmt, st, level + 2, tableInfo);
            }
        }

        void columnsInTableToken(TCustomSqlStatement stmt, TSourceToken st, int level, tableInfo tableInfo)
        {

            if (st.RelatedToken != null)
            {
                // declared table alias token
                TSourceToken rt = st.RelatedToken;
                TSourceToken rrt = null;
                for (int i = 0; i < rt.RelatedTokens.Count(); i++)
                {
                    rrt = rt.RelatedTokens[i];
                    if (rrt.ChildToken != null)
                    {
                        columnInfo columnInfo = new columnInfo(tableInfo);
                        columnInfo.column = rrt.ChildToken;
                        columnInfos.Add(columnInfo);
                    }
                }
            }

            TSourceToken rtt = null;
            for (int i = 0; i < st.RelatedTokens.Count(); i++)
            {
                // reference table token
                rtt = st.RelatedTokens[i];
                if (rtt.DBObjType == TDBObjType.ttObjField)
                {
                    // get all field tokens link with table token (those token not linked by syntax like tablename.fieldname)
                    // but like this : select f from t

                    columnInfo columnInfo = new columnInfo(tableInfo);
                    columnInfo.column = rtt;
                    columnInfos.Add(columnInfo);
                }
                if (rtt.ChildToken != null)
                {
                    columnInfo columnInfo = new columnInfo(tableInfo);
                    columnInfo.column = rtt.ChildToken;
                    columnInfos.Add(columnInfo);
                }
            }
        }

        void tablesInStmt(TCustomSqlStatement stmt, int level)
        {

            tableTokensInStmt(stmt, level);

            for (int j = 0; j < stmt.ChildNodes.Count(); j++)
            {
                if (stmt.ChildNodes[j] is TCustomSqlStatement)
                {
                    tablesInStmt(stmt.ChildNodes[j] as TCustomSqlStatement, level + 1);
                }
            }

        }
    }
}
